@@ -31,25 +31,43 @@ struct pv_area {
 	uint32_t start;
 	uint32_t count;
 
-	struct list list;		/* pv_map.areas */
+	/* Number of extents unreserved during a single allocation pass. */
+	uint32_t unreserved;
+
+	struct dm_list list;		/* pv_map.areas */
+};
+
+/*
+ * When building up a potential group of "parallel" extent ranges during
+ * an allocation attempt, track the maximum number of extents that may
+ * need to be used as a particular parallel area.  Several of these
+ * structs may reference the same pv_area, but 'used' may differ between
+ * them.  The sum of all the 'used' variables referring to the same
+ * pv_area may not exceed that area's count, so we cannot allocate the
+ * same extents twice.
+ */
+struct pv_area_used {
+	struct pv_area *pva;
+	uint32_t used;
 };
 
 struct pv_map {
 	struct physical_volume *pv;
-	struct list areas;		/* struct pv_areas */
+	struct dm_list areas;		/* struct pv_areas */
 	uint32_t pe_count;		/* Total number of PEs */
 
-	struct list list;
+	struct dm_list list;
 };
 
 /*
  * Find intersection between available_pvs and free space in VG
  */
-struct list *create_pv_maps(struct dm_pool *mem, struct volume_group *vg,
-			    struct list *allocatable_pvs);
+struct dm_list *create_pv_maps(struct dm_pool *mem, struct volume_group *vg,
+			    struct dm_list *allocatable_pvs);
 
 void consume_pv_area(struct pv_area *area, uint32_t to_go);
+void reinsert_changed_pv_area(struct pv_area *pva);
 
-uint32_t pv_maps_size(struct list *pvms);
+uint32_t pv_maps_size(struct dm_list *pvms);
 
 #endif
